import regex as re
from typing import List

def has_valid_chars(entity: str, allow_cyrillic: bool = True) -> bool:
    """ Checks if entity contains any valid characters in latin
    or in cyrillic, if the latter is enabled

    Parameters
    ------------
    entity: str
        String to validate.
    allow_cyrillic: bool
        Allow strings in cyrillic?

    Returns
    ------------
    bool
        Boolean value indicating, if the string
        contains any valid characters.

    """
    # Check for latin characters
    is_valid = bool(re.search(r"[a-züõöäA-ZÜÕÖÄ]", entity))

    if allow_cyrillic and not is_valid:
        # If cyrillic characters are allowed,
        # check for them as well
        is_valid = bool(re.search(r"[а-яА-Я]", entity))

    return is_valid


def filter_names(names: List[str], allow_cyrillic: bool = True) -> List[str]:
    """ Filters out names not in allowed encodings (latin / cyrillic).

    Parameters
    ------------
    names: List[str]
        Names to filters.
    allow_cyrillic: bool
        Allow strings in cyrillic?

    Returns
    ------------
    List[str]
        List of filtered names.

    """
    filtered_names = [
        name for name in names
        if has_valid_chars(entity=name, allow_cyrillic=allow_cyrillic)
    ]
    return filtered_names
    


import os

from rara_tools.task_reporter import TaskReporter

REPORTER = TaskReporter(
    os.getenv("TASK_REPORTER_URL", "http://localhost:8000/api/v1/"),
    os.getenv("TASK_REPORTER_TOKEN", "")
)

# TODO: write tests when Core API is ready.

# def test_api_health():
#     """Tests TaskReporter's health check.
#     """
#     # normal response
#     response = REPORTER.check()
#     assert response is True
#     # wrong URL
#     response = TaskReporter("http://whatever.com", "xxx").check()
#     assert response is False

# def test_api_report():
#     """Tests reporting functionaly of the class.
#     """
#     data_to_patch = {
#         "number_of_pages": 10,
#         "elasticID": "foobar"
#     }
#     task_id = 1
#     response = REPORTER.report_results(data_to_patch, task_id)
#     print(response)



from typing import List, Tuple, Any
from rara_tools.core_formatters.formatted_keyword import FormattedKeyword
from rara_tools.core_formatters.formatted_meta import FormattedAuthor
from rara_tools.constants.linker import MAIN_TAXONOMY_LANG, KEYWORD_TYPES_TO_IGNORE, EntityType
from rara_tools.constants.meta_extractor import TitleType, TITLE_TYPES_MAP, PUBLISHER_KEY
from rara_tools.constants.subject_indexer import KeywordType

import regex as re

def get_primary_author(authors: List[dict]) -> str:
    primary_author = ""
    for author in authors:
        if author.get("is_primary", False):
            primary_author = author.get("name", "")
    return primary_author

def is_valid_keyword(keyword: str) -> bool:
    # If keywords contains ONLY punctuation
    # characters, we assume it`s not valid
    if re.search(r"^(\W|_)+$", keyword):
        return False
    return True

def format_series_info(series: str):
    pass

def format_authors(authors: List[dict]) -> Tuple[List[dict], dict]:
    formatted_authors = []
    publisher = {}
    for author in authors:
        entity_type = author.get("type", EntityType.UNK)

        formatted_author = FormattedAuthor(
            object_dict=author,
            linked_doc=None,
            entity_type=entity_type
        ).to_dict()

        # If author role == publisher, do not add it as an author
        if formatted_author.get("author_role", "") == PUBLISHER_KEY:
            publisher = formatted_author
            continue

        formatted_authors.append(formatted_author)
    return (formatted_authors, publisher)

def format_sections(sections: List[dict]) -> List[dict]:
    for section in sections:
        authors = section.pop("authors", [])
        titles = section.pop("titles", [])
        primary_author = get_primary_author(authors)
        if primary_author:
            for title in titles:
                title["author_from_title"] = primary_author
        if not authors:
            for title in titles:
                title["title_type"] = TitleType.ANON
                title["title_type_int"] = TITLE_TYPES_MAP.get(TitleType.ANON)
        section["titles"] = titles

        # Extract publisher, but do nothing with it
        # as it is unlikely for the publishing info to be
        # in a METS/ALTO section. Can update it, if proven otherwise
        formatted_authors, publisher = format_authors(authors)
        section["authors"] = formatted_authors

    return sections

def format_meta(meta: dict) -> dict:
    """ Formats unlinked meta for Kata CORE.
    """

    meta_to_format = meta.get("meta")

    authors = meta_to_format.pop("authors", [])
    sections = meta_to_format.pop("sections", [])
    titles = meta_to_format.pop("titles", [])

    formatted_authors, publisher = format_authors(authors)
    formatted_sections = format_sections(sections)

    if sections and formatted_sections:
        meta_to_format["sections"] = formatted_sections
    if authors and formatted_authors:
        meta_to_format["authors"] = formatted_authors
    if titles and not authors:
        for title in titles:
            title["title_type"] = TitleType.ANON
            title["title_type_int"] = TITLE_TYPES_MAP.get(TitleType.ANON)
        meta_to_format["titles"] = titles

    if publisher:
        # Not sure, if it would be better to add original name or
        # linked value. Currently adding original for safety
        meta_to_format["publisher"] = publisher.get("original_name")

    meta["meta"] = meta_to_format

    return meta


def format_keywords(flat_keywords: List[dict]) -> List[dict]:
    """ Formats unlinked keywords for Kata CORE.
    """
    ignored_keywords = []
    filtered_keywords = []

    for keyword_dict in flat_keywords:
        keyword_type = keyword_dict.get("entity_type")
        if keyword_type in KEYWORD_TYPES_TO_IGNORE:
            ignored_keywords.append(keyword_dict)
        else:
            filtered_keywords.append(keyword_dict)

    formatted_keywords = []

    for keyword_dict in filtered_keywords:
        formatted_keyword = FormattedKeyword(
            object_dict=keyword_dict,
            linked_doc=None,
            main_taxnomy_lang=MAIN_TAXONOMY_LANG
        ).to_dict()
        if is_valid_keyword(formatted_keyword.get("keyword")):
            formatted_keywords.append(formatted_keyword)

    return formatted_keywords

def get_udk072(flat_keywords: List[dict]) -> List[str]:
    """ Filters out UDK from flat subject indexer output.
    """
    # keyword type: UDK
    udk072 = [
        keyword.get("keyword")
        for keyword in flat_keywords
        if keyword.get("entity_type") == KeywordType.UDK
    ]
    return udk072


def get_udk080(flat_keywords: List[dict]) -> List[str]:
    """ Filters out UDC from flat subject indexer output.
    """
    # keyword type: UDC
    udk080 = [
        keyword.get("keyword")
        for keyword in flat_keywords
        if keyword.get("entity_type") == KeywordType.UDC
    ]
    return udk080



from pymarc import Field, Subfield, Record
from typing import List, Optional

from rara_tools.constants import EMPTY_INDICATORS
from rara_tools.normalizers.viaf import VIAFRecord
from rara_tools.normalizers import RecordNormalizer
from rara_tools.normalizers.exceptions import SkipRecordException


import logging


logger = logging.getLogger(__name__)


class BibRecordNormalizer(RecordNormalizer):
    """Normalize bib records."""

    def __init__(
        self,
        linking_results: List[dict] = [],
        sierra_data: List[dict] = [],
        classified_fields: List[List[dict]] = [],
        ALLOW_EDIT_FIELDS: List[str] = ["008", "925"],
        REPEATABLE_FIELDS: List[str] = ["667"],
    ):
        super().__init__(linking_results, sierra_data, classified_fields)
        self.DEFAULT_LEADER = "00399nz  a2200145n  4500"  # must be 24 digits
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS

        self.records_extra_data = []
        self.sierra_data = sierra_data
        self.records = self._setup_records(
            linking_results, sierra_data, classified_fields
        )

    def _normalize_sierra(
        self, record: Record, is_editing_existing_record: bool
    ) -> Record:
        suffix_008 = "|||aznnnaabn          || |||      "

        if is_editing_existing_record:
            # Try to reuse prefix from existing 008 field if present
            existing_008 = next(
                (f for f in record.fields if f.tag == "008" and hasattr(f, "data")),
                None,
            )
            if existing_008 and len(existing_008.data) >= 6:
                prefix = existing_008.data[:6]  # keep existing timestamp
            else:
                prefix = self.current_timestamp()  # fallback if no valid existing data
        else:
            prefix = self.current_timestamp()

        fields = [
            Field(tag="008", data=f"{prefix}{suffix_008}"),
        ]

        self._add_fields_to_record(record, fields)

    def _add_author(
        self, record: Record, viaf_record: Optional[VIAFRecord], original_entity: str
    ) -> Optional[Field]:
        if record.get("100") or record.get("110") or record.get("111"):
            return record

        type_map = {"Personal": "100", "Corporate": "110", "Collective": "111"}

        tag = type_map.get(getattr(viaf_record, "name_type", None), "100")

        # KRAT-785 if author cannot be found at all
        title = getattr(viaf_record, "name", None)
        if not title:
            logger.info(
                f"No author name found in classified data or VIAF for original entity: {original_entity}, skipping creating bib normrecord"
            )
            # Skip to the next iteration
            raise SkipRecordException("No author name found, skipping record")

        # Note that without title, the record will not be created further along in the pipeline
        fields = [
            Field(
                tag=tag, indicators=EMPTY_INDICATORS, subfields=[Subfield("t", title)]
            )
        ]

        if viaf_record:
            author_dates = self.get_formatted_dates(viaf_record)
            if author_dates:
                fields[0].add_subfield("d", author_dates)

        self._add_fields_to_record(record, fields)

        if viaf_record:
            self._include_name_variations(record, viaf_record, author_tag=tag)

    def _normalize_viaf(
        self, record: Record, viaf_record: VIAFRecord, original_entity: str
    ) -> None:
        if not viaf_record:
            # viaf record not found, include original entity as 100|t
            self._add_author(record, viaf_record=None, original_entity=original_entity)
            return record

        self._add_author(record, viaf_record, original_entity=original_entity)

    def _normalize_record(
        self,
        record: Record,
        sierraID: str,
        viaf_record: VIAFRecord,
        is_editing_existing_record: bool,
        original_entity: str,
    ) -> Record:
        self._normalize_sierra(record, is_editing_existing_record)
        self._normalize_viaf(record, viaf_record, original_entity=original_entity)

        return record



from tests.const import SIERRA_INPUT_DIR, LINKER_DIR

from rara_tools.converters import SierraResponseConverter
from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient

from pymarc import Record
from typing import List

import json
import os


def read_json_file(path: str):
    with open(path, "r") as f:
        data = f.read()
        return json.loads(data)


def check_record_tags_sorted(record: Record):
    record_tags = [field.tag for field in record.get_fields()]
    assert record_tags == sorted(record_tags)


def check_no_dupe_tag_values(record: Record):
    repetable_tags = ["024", "035", "400", "670"]
    record_tags = [field.tag for field in record.get_fields()
                   if field.tag not in repetable_tags]
    assert len(record_tags) == len(set(record_tags))


def check_record_tags_have_values(record: Record, tags: List[str]):
    for tag in tags:
        assert record[tag] is not None


def get_record_field_value(record: Record, tag: str):
    """ handle control & variable fields """
    return record.get_fields(tag)[0].value()


def compare_results(expected: dict, results: dict):
    return json.dumps(expected) == json.dumps(results)


def get_formatted_sierra_response(fname: str):
    """ Reads a mock Sierra response file and converts it to MARC in json."""

    response = read_json_file(os.path.join(SIERRA_INPUT_DIR, fname))

    converter = SierraResponseConverter(response)
    return converter.convert()


def get_viaf_record(id: str, allowed_sources: list):
    """ Fetches VIAF record by ID and returns a VIAFRecord object """

    client = VIAFClient()
    response = client.get_records_by_viaf_id(id)

    viaf_record = VIAFRecord(
        response, allowed_sources=allowed_sources)
    return viaf_record


def search_viaf_record(search_term: str, allowed_sources: list):
    """ Fetches VIAF record by name and returns a VIAFRecord object """
    client = VIAFClient()
    response = client.get_records_by_search_term(search_term)

    return VIAFRecord(response, allowed_sources=allowed_sources)


def get_linker_res_example(fname: str):
    with open(os.path.join(LINKER_DIR, fname), "r") as f:
        data = f.read()
        return json.loads(data)



from pymarc import Indicators

YYMMDD_FORMAT = "%y%m%d"
YYYYMMDD_FORMAT = "%Y%m%d"
YY_DD_FORMAT = "%Y-%m"

class EntityType:
    PER = "PER"
    ORG = "ORG"
    KEYWORD = "EMS_KEYWORD"
    LOC = "LOC"
    TITLE = "TITLE"
    UNK = "UNKNOWN"


EMPTY_INDICATORS = Indicators(" ", " ")
VIAF_ALLOWED_SOURCES = ["LC", "DNB", "LNB", "NLL",
                        "ERRR", "J9U"]

DEFAULT_VIAF_FIELD = "local.names"

ALLOWED_VIAF_FIELDS = [
    "cql.any",                          # All fields
    "local.names",                      # All headings
    "local.personalNames",              # Personal names
    "local.corporateNames",             # Corporate names
    "local.geographicNames",            # Geographic names
    "local.uniformTitleWorks",          # Works
    "local.uniformTitleExpressions",    # Expressions
    "local.mainHeadingEl",              # Preferred headings
    "Xlocal.names",                     # Exact headings
    "local.title"                       # Bibliographic titles
]

# For mapping rara-linker's entity type's to corresponding VIAF fields
VIAF_ENTITY_MAP = {
    EntityType.PER: "local.personalNames",
    EntityType.ORG: "local.corporateNames",
    EntityType.LOC: "loca.geographicNames",
    EntityType.TITLE: "local.uniformTitleWorks"


}
ALLOWED_VIAF_WIKILINK_LANGS = ["en", "et"]
VIAF_SIMILARITY_THRESHOLD = 0.92
VERIFY_VIAF_RECORD = True
MAX_VIAF_RECORDS_TO_VERIFY = 10



from datetime import datetime
from pymarc import Field, Subfield, JSONReader, Record
from typing import List, Optional, Iterator
from rara_tools.normalizers.reader import SafeJSONReader

from rara_tools.parsers.tools.validators import filter_names
from rara_tools.normalizers.exceptions import SkipRecordException


from rara_tools.normalizers.viaf import VIAFRecord, VIAFClient
from rara_tools.constants.normalizers import (
    DEFAULT_VIAF_FIELD,
    ALLOWED_VIAF_FIELDS,
    ALLOWED_VIAF_WIKILINK_LANGS,
    VIAF_SIMILARITY_THRESHOLD,
    VERIFY_VIAF_RECORD,
    MAX_VIAF_RECORDS_TO_VERIFY,
    EMPTY_INDICATORS,
    YYMMDD_FORMAT,
    YY_DD_FORMAT,
    YYYYMMDD_FORMAT,
)
from glom import glom
from dateutil import parser
from datetime import date

import logging
import json

logger = logging.getLogger(__name__)


class RecordNormalizer:
    """
    Base class for normalizing different record types corresponding classes have been created.
    By default existing record fields will not be changed, unless included in ALLOW_EDIT_FIELDS. If a field
    included in the normalization is not present, it will be added to the record. If under REPEATABLE_FIELDS.
    a new record field is added.

    Args:
        sierra_data: Optionally, can normalize records from SIERRA. Must be in specific format,
        e.g converted with SierraResponseConverter. examples at: tests/sierra/output
        classified_fields: Optionally can include marc fields, will follow the rules of the tag number.
        Useful to send classified data from core.
        entities: List of Full names (str). If included, will use NormLinker to match with normalized records on KATA elastic.
    """

    def __init__(
        self,
        linking_results: List[dict] = [],
        sierra_data: List[dict] = [],
        classified_fields: List[List[dict]] = [],
        ALLOW_EDIT_FIELDS: List[str] = ["925"],
        REPEATABLE_FIELDS: List[str] = ["667"],
    ):
        # Include, if will replace existing field
        self.ALLOW_EDIT_FIELDS = ALLOW_EDIT_FIELDS
        # include, if should be added alongside existing fields
        self.REPEATABLE_FIELDS = REPEATABLE_FIELDS
        # leader applied to new records
        self.DEFAULT_LEADER = "01682nz  a2200349n  4500"  # must be 24 digits

    def _setup_records(
        self,
        linking_results: List[dict],
        sierra_data: List[dict],
        classified_fields: List[List[dict]] = [],
    ) -> JSONReader:
        """Setup initial MARC records and data.

        If no linked entities or more than one linked entity found, we create a new record.
        If one linked entity found, we create an updated record from the linked entity data.
        """
        linked_records = []

        def handle_create_new_record(entity, idx, elastic_meta):
            logger.info(f"No linked entities found for {entity}, Creating new record.")
            linked_records.append({"leader": self.DEFAULT_LEADER, "fields": []})
            self.records_extra_data.append(
                {
                    "entity": entity,
                    "classified_fields": classified_fields[idx]
                    if idx < len(classified_fields)
                    else [],
                    "edited": False,
                    "elastic": elastic_meta,
                }
            )

        for idx, linked in enumerate(linking_results or []):
            if not isinstance(linked, dict):
                logger.debug(f"Skipping invalid linked result: {linked}")
                continue

            entity = linked.get("original_entity")
            linked_info = linked.get("linked_info", [])

            if not isinstance(linked_info, list) or not linked_info:
                # No linked entities found, create new record
                elastic_meta = {}  # Always empty in this case
                handle_create_new_record(entity, idx, elastic_meta)
                continue

            elif len(linked_info) > 1:
                # Multiple linked entities found, Skip
                logger.info(
                    f"Multiple ({len(linked_info)}) linked entities found for {entity}, skipping creating normrecord."
                )
                continue

            elif len(linked_info) == 1:
                # one record match found, we update existing record

                linked_item = linked_info[0]
                if not isinstance(linked_item, dict):
                    continue

                elastic_meta = linked_item.get("elastic", {})

                # handle case where we have linked an entity without a record
                if not linked_item.get("json", None):
                    handle_create_new_record(entity, idx, elastic_meta)
                    continue

                linked_records.append(linked_item.get("json", {}))

                self.records_extra_data.append(
                    {
                        "entity": entity,
                        "viaf": linked_item.get("viaf", {}),
                        "classified_fields": classified_fields[idx]
                        if idx < len(classified_fields)
                        else [],
                        "edited": True,
                        "elastic": elastic_meta,
                    }
                )
                continue

        self.records_extra_data.extend(
            {"sierraID": obj.get("sierraID"), "edited": True}
            for obj in (sierra_data or [])
            if isinstance(obj, dict)
        )

        all_records = linked_records + (sierra_data or [])

        return SafeJSONReader(json.dumps(all_records, ensure_ascii=False))

    @staticmethod
    def current_timestamp():
        """6 digit timestamp, format YYMMDD"""
        return datetime.now().strftime(YYMMDD_FORMAT)

    @staticmethod
    def current_yyyy_dd():
        """format of 2025-03"""
        return datetime.now().strftime(YY_DD_FORMAT)

    @staticmethod
    def _is_person_est_nationality(viaf_record: VIAFRecord) -> bool:
        return hasattr(viaf_record, "nationality") and viaf_record.nationality == "ee"

    def get_formatted_dates(self, viaf_record: VIAFRecord) -> str | None:
        """Get birth and death date in the form 1878-1940. If only birth date is present, return 1878-.
        If no dates, return empty string.
        """
        birth_date = self._extract_year(viaf_record.birth_date)
        death_date = (
            self._extract_year(viaf_record.death_date)
            if viaf_record.death_date != 0
            else ""
        )

        if birth_date and death_date:
            return f"{birth_date}-{death_date}"
        elif birth_date:
            return f"{birth_date}-"
        else:
            return None

    def _is_nxx(self, field: Field, n: str):
        """Check if fields tag is in nxx range."""
        return field.tag.startswith(n)

    def get_record_field_or_none(self, record: Record, tag: str) -> Optional[Field]:
        return record.get_fields(tag)[0] if record.get_fields(tag) else None

    def _field_in_record(self, field: Field, record: Record) -> bool:
        """Check if field exists in record."""
        existing_fields = record.get_fields(field.tag)
        return any(
            field.data == existing_field.data for existing_field in existing_fields
        )

    def _filter_equivalent_field_not_in_record(
        self, record: Record, fields: List[Field]
    ) -> bool:
        """filter out fields, that do not have an equivalent in the record."""
        return filter(lambda field: not self._field_in_record(field, record), fields)

    def _extract_year(self, value: str) -> str:
        if value is None:
            return ""

        if isinstance(value, (datetime, date)):
            return str(value.year)

        try:
            dt = parser.parse(str(value), fuzzy=True)
            parsed_year = str(dt.year)
            logger.info(f"Extracted year '{parsed_year}' from value '{value}'")
            return parsed_year
        except Exception as e:
            logger.info(f"Failed to extract year string '{value}': {e}")
            return ""

    def _format_date(self, value: str) -> str:
        if not value:
            return ""

        if isinstance(value, (datetime, date)):
            return value.strftime(YYYYMMDD_FORMAT)

        val = str(value).strip()

        try:
            dt = parser.parse(val, fuzzy=False, default=datetime(1, 1, 1))
        except Exception:
            return ""

        if len(val) == 4 and val.isdigit():
            return dt.strftime("%Y")  # YYYY
        if len(val) in (6, 7):  # YYYYMM or YYYY-MM
            return dt.strftime("%Y%m")  # YYYYMM
        return dt.strftime(YYYYMMDD_FORMAT)  # YYYYMMDD

    def get_subfield(
        self, record: Record, tag: str, subfield: str, default: str
    ) -> str:
        """get record existing subfield value or assign a fallback value."""

        field = self.get_record_field_or_none(record, tag)

        if field is None:
            return default

        subfields = field.get_subfields(subfield)
        return subfields[0] if subfields else default

    def _handle_default_fields(self, record: Record, *fields: List[Field]) -> Record:
        """Default behavior - add field to record iff not present already"""
        record.add_field(
            *filter(
                lambda field: field.tag not in [f.tag for f in record.get_fields()],
                fields,
            )
        )

    def _handle_editable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """replace existing field with a new field."""

        editable_fields = filter(
            lambda field: field.tag in self.ALLOW_EDIT_FIELDS, fields
        )

        tags = [f.tag for f in editable_fields]

        record.remove_fields(*tags)
        record.add_field(*editable_fields)

    def _handle_repeatable_fields(self, record: Record, *fields: List[Field]) -> Record:
        """add field to the record & don't replace existing field."""

        repeatable_fields = [
            field for field in fields if field.tag in self.REPEATABLE_FIELDS
        ]

        record.add_field(
            *repeatable_fields
            # *self._filter_equivalent_field_not_in_record(
            #     record, repeatable_fields)
        )

    def _add_fields_to_record(self, record: Record, fields: List[Field]) -> Record:
        cleaned_fields = []

        for field in fields:
            # Always assume control fields cleaned
            if field.tag < "010" and field.tag.isdigit():
                cleaned_fields.append(field)
                continue

            # filter out subfields that are empty or 0 (VIAF returns 0 for unknown dates)
            field.subfields = [
                sub
                for sub in field.subfields
                if sub.value and sub.value not in ["0", 0]
            ]

            # only keep the field if it still has subfields left
            if field.subfields:
                cleaned_fields.append(field)

        if not cleaned_fields:
            return record

        self._handle_repeatable_fields(record, *cleaned_fields)
        self._handle_editable_fields(record, *cleaned_fields)
        self._handle_default_fields(record, *cleaned_fields)

        return record

    def _add_author(self, record: Record, viaf_record: VIAFRecord) -> Optional[Field]:
        existing_author: Optional[Field] = (
            record.get("100") or record.get("110") or record.get("111")
        )

        if existing_author:
            # Attempt to add alternative name variations only
            # if viaf_record:
            #     author_tag = hasattr(existing_author, 'tag') and existing_author.tag or "100"
            #     self._include_name_variations(record, viaf_record, author_tag=author_tag)
            return record

        type_map = {"Personal": "100", "Corporate": "110", "Collective": "111"}
        author_type = viaf_record.name_type
        tag = type_map.get(author_type, "100")

        fields = [
            Field(
                tag=tag,
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("a", viaf_record.name),
                ],
            )
        ]

        if viaf_record:
            author_dates = self.get_formatted_dates(viaf_record)
            if author_dates:
                fields[0].add_subfield("d", author_dates)

        self._add_fields_to_record(record, fields)

        if viaf_record:
            self._include_name_variations(record, viaf_record, author_tag=tag)

    def _include_name_variations(
        self,
        record: Record,
        viaf_record: VIAFRecord,
        author_tag: str,
        filter_variations=True,
    ) -> None:
        """Include name variations from VIAF record as 400|t fields"""

        if not viaf_record or not viaf_record.name_variations:
            return

        existing_name_variations = record.get_fields("400")
        existing_variations = [
            sf.value
            for field in existing_name_variations
            for sf in field.get_subfields("a")
        ]

        if filter_variations:
            allowed_variations = filter_names(viaf_record.name_variations)
            logger.debug(
                f"filtered out {len(viaf_record.name_variations) - len(allowed_variations)} name variations for '{viaf_record.name}'"
            )

        else:
            allowed_variations = viaf_record.name_variations

        fields = []

        tag = "4" + author_tag[1:] if author_tag in ["100", "110", "111"] else "400"

        for variation in allowed_variations:
            if variation not in existing_variations:
                fields.append(
                    Field(
                        tag=tag,
                        indicators=EMPTY_INDICATORS,
                        subfields=[Subfield("a", variation)],
                    )
                )

        self._add_fields_to_record(record, fields)

    def _move680_fields_to_667(self, record: Record) -> None:
        """Move existing 680 fields to 667, if any."""
        fields_680 = record.get_fields("680")
        if not fields_680:
            return

        fields_667 = [
            Field(tag="667", indicators=EMPTY_INDICATORS, subfields=field.subfields)
            for field in fields_680
        ]

        record.remove_fields("680")
        self._add_fields_to_record(record, fields_667)

    def _include_classified_fields(
        self, record: Record, classified_fields: list[dict]
    ) -> None:
        """Include classified fields from core, if any.
        e.g. classified_fields=[{'670': {'ind1': ' ', 'ind2': '0', 'subfields': [{'a': 'Päikesekiri, 2021'}]}}]

        For each record, we need a list of dicts, to handle repeatable fields.
        """
        if not classified_fields:
            return

        fields = [
            Field(
                tag=str(tag),
                indicators=v.get(
                    "indicators", [v.get("ind1", " "), v.get("ind2", " ")]
                ),
                subfields=[
                    Subfield(code, value)
                    for sub in v.get("subfields", [])
                    for code, value in sub.items()
                ],
            )
            for field_dict in classified_fields
            for tag, v in field_dict.items()
        ]

        self._add_fields_to_record(record, fields)

    def _normalize_common(
        self,
        record: Record,
        is_editing_existing_record: bool,
        classified_fields: List[dict],
    ) -> None:
        """Common logic for all normalizations.
        - Includes note about record being created/edited.
        - include date note with a different subfield, depending on if record is new or edited.
        - move existing 680 fields to 667
        """
        self._include_classified_fields(record, classified_fields)

        # before adding new notes
        self._move680_fields_to_667(record)

        note = "Muudetud AI poolt" if is_editing_existing_record else "Loodud AI poolt"
        date_note = f"KRATT {self.current_yyyy_dd()}"

        field_667 = Field(
            tag="667", indicators=EMPTY_INDICATORS, subfields=[Subfield("a", note)]
        )

        fields = [field_667]

        if is_editing_existing_record:
            field_925 = Field(
                tag="925",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("p", self.get_subfield(record, "925", "p", date_note))
                ],
            )
            fields.append(field_925)

        else:
            field_925 = Field(
                tag="925",
                indicators=EMPTY_INDICATORS,
                subfields=[
                    Subfield("t", self.get_subfield(record, "925", "t", date_note))
                ],
            )
            fields.append(field_925)

        self._add_fields_to_record(record, fields)

        return record

    def _get_viaf_search_term(
        self, record: Record, entity: Optional[str]
    ) -> Optional[str]:
        """prioritize entity name, if not available, use author name."""
        if entity:
            return entity

        author_field = record.get("100") or record.get("110") or record.get("111")
        if author_field:
            return (
                author_field.get_subfields("a")[0]
                if author_field.get_subfields("a")
                else None
            )

        logger.warning(
            "No entity or author name found for VIAF search. Skipping VIAF enrichment."
        )

    def _get_viaf_record(
        self,
        record: Record,
        is_editing_existing_record: bool,
        viaf_id: Optional[int] = None,
        entity: Optional[str] = None,
        viaf_identifier: Optional[str] = None,
        viaf_field: str = DEFAULT_VIAF_FIELD,
        threshold: float = VIAF_SIMILARITY_THRESHOLD,
        verify: bool = VERIFY_VIAF_RECORD,
        max_records: int = MAX_VIAF_RECORDS_TO_VERIFY,
    ) -> Optional[VIAFRecord]:
        """

        Get VIAF record either by ID if possible, otherwise by search term.
        We do not use the entity as the search term directly, due to inaccuracies.

        Instead, search by search term user only if viaf_identifier is available

        """

        viaf_record = None

        try:
            viaf_client = VIAFClient()

            if viaf_id:
                viaf_records = viaf_client.get_normalized_data_by_ids([viaf_id])
                if viaf_records:
                    viaf_record = viaf_records[0]
            else:
                # KRATT-787
                #  If editing, only allow to use viaf_identifier
                if is_editing_existing_record:
                    search_term = viaf_identifier if viaf_identifier else None
                else:
                    search_term = self._get_viaf_search_term(record, entity)
                    verify = True  # more accurate with author name search

                if search_term:
                    logger.info(
                        f"Searching for VIAF record with search term: {search_term}"
                    )

                    if not verify:
                        logger.warning(
                            f"Record verification is turned off. If multiple records are "
                            f"detected for search term '{search_term}', the first "
                            f"result is automatically returned. This might lead to "
                            f"some inaccuracies!"
                        )

                    viaf_record = viaf_client.get_normalized_data_by_search_term(
                        search_term=search_term,
                        field=viaf_field,
                        max_records=max_records,
                        verify=verify,
                        threshold=threshold,
                    )
                    if viaf_record:
                        logger.debug(
                            f"VIAF {search_term}, linked to ID: {viaf_record.viaf_id}"
                        )

        except Exception as e:
            logger.error(
                f"Error fetching VIAF record with ID={viaf_id} / entity='{entity}': {e}"
            )
        return viaf_record

    def _normalize_record(
        self,
        record: Record,
        sierraID: str,
        viaf_record: VIAFRecord,
        is_editing_existing_record: bool,
        original_entity: str,
    ) -> Record:
        return record

    def get_record(self, index: int) -> Record:
        """Get normalized record by index."""
        for idx, record in enumerate(self):
            if idx == index:
                return record
        raise IndexError("Record index out of range.")

    @property
    def data(self) -> List[dict]:
        """Shorthand to get all normalized records as dict, skipping failures."""
        result = []
        for record in self:
            try:
                result.append(record.as_dict())
            except Exception as e:
                logger.error(f"Failed to normalize record: {e}")
                continue
        return result

    @property
    def first(self) -> Record:
        return next(iter(self))

    def __iter__(self) -> Iterator:
        # viaf_id_path = "viaf.original.queryResult.viafID"
        viaf_id_path = "viaf.parsed.viaf_id"
        viaf_identifier_path = "elastic.identifier"
        sierra_id_path = "sierraID"

        for record, extra_data in zip(self.records, self.records_extra_data):
            sierra_id = glom(extra_data, sierra_id_path, default="")
            viaf_id = glom(extra_data, viaf_id_path, default=None)
            # KRAT-787 - In our elasticsearch docs, we store an "identifier"
            # field, that is infrequently filled, but can be used for exact VIAF record search.
            viaf_identifier = glom(extra_data, viaf_identifier_path, default=None)
            logger.info(
                f"Processing record with VIAF ID: {viaf_id}, VIAF Identifier: {viaf_identifier}"
            )

            classified_fields = extra_data.get("classified_fields", [])
            entity = extra_data.get("entity")
            is_editing_existing_record = extra_data.get("edited") is True

            viaf_record = self._get_viaf_record(
                record=record,
                is_editing_existing_record=is_editing_existing_record,
                viaf_id=viaf_id,
                entity=entity,
                viaf_identifier=viaf_identifier,
                verify=False,  # Not accurate with identifier search
            )
            if viaf_record:
                logger.debug(
                    f"linked VIAF record with ID {viaf_record.viaf_id} for entity '{entity}'"
                )

            record = self._normalize_common(
                record, is_editing_existing_record, classified_fields
            )
            
            try:
                normalized_record = self._normalize_record(
                    record,
                    sierra_id,
                    viaf_record,
                    is_editing_existing_record,
                    original_entity=entity,
                )
            except SkipRecordException as e:
                logger.info(f"Skipping record normalization: {e}")
                continue

            normalized_record.fields.sort(key=lambda field: field.tag)

            yield normalized_record



from iso639 import Lang
from datetime import datetime


def lang_to_iso639_1(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-1 standard.
    Input can be any language code in a valid ISO-639
    standard or even a full name of the language,
    e.g. "Estonian".

    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.

    unk_code: str
        Code to return incase of invalid/unsupported
        input language.

    Returns
    -------
    Language code in ISO-639-1 standard.
    """
    try:
        lg = Lang(lang)
        iso_639_1_lang = lg.pt1
    except:
        iso_639_1_lang = unk_code
    return iso_639_1_lang


def lang_to_iso639_2(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-2 standard.
    Input can be any language code in a valid ISO-639
    standard or even a full name of the language,
    e.g. "Estonian".

    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.

    unk_code: str
        Code to return incase of invalid/unsupported
        input language.

    Returns
    -------
    Language code in ISO-639-2 standard.
    """
    try:
        lg = Lang(lang)
        # NB! uses bibliographic identifier (e.g. "de" -> "ger")
        # opposed to terminological identifier ("de" -> "deu").
        # This can be changed by replaving lg.pt2b -> lg.pt2t
        iso_639_2_lang = lg.pt2b
    except:
        iso_639_2_lang = unk_code
    return iso_639_2_lang


def lang_to_iso639_3(lang: str, unk_code: str = "unk") -> str:
    """ Converts language into ISO-639-3 standard.
    Input can be any language code in a valid ISO-639
    standard or even a full name of the language,
    e.g. "Estonian".

    Parameters
    -----------
    lang: str
        Language code in any valid ISO-639 standard.
        unk_code: str

    Code to return incase of invalid/unsupported
        input language.

    Returns
    -------
    str
        Language code in ISO-639-3 standard.
    """
    try:
        lg = Lang(lang)
        iso_639_3_lang = lg.pt3
    except:
        iso_639_3_lang = unk_code
    return iso_639_3_lang


def ratio_to_percentage(ratio: float) -> str:
    """ Converts ratio to corresponding percentage.

    Parameters
    -----------
    ratio: float
        Float in range [0,1]

    Returns
    --------
    str
        Percentage corresponding to the float.

    """
    percentage = f"{int(ratio*100)}%"
    return percentage

def format_date(original_date: str) -> str:
    """ Converts date from format %Y-%m-%d into format %d.%m.%Y, e.g:
    2025-02-12 -> 12.02.2025.

    Parameters
    -----------
    original_date: str
        Original date in format %Y-%m-%d

    Returns
    ----------
    str:
        Date in format %d.%m.%Y
    """
    try:
        date_obj = datetime.strptime(original_date, "%Y-%m-%d")
        #new_date = date_obj.strftime("%d.%m.%Y") # Would have been more precise :(
        new_date = date_obj.strftime("%Y")

    except:
        new_date = original_date
    return new_date



import logging

LOGGER = logging.getLogger("rara-tools-marc-parser")


class EMSMarcIDs:
    SYNONYMS = ["448", "450", "451", "455"]
    RELATED = ["548", "550", "551", "555"]
    CATEGORY = ["072"]
    NOTES = ["680"]
    URL = ["024"]
    TIME_KEYWORD = "148"
    TOPIC_KEYWORD = "150"
    LOC_KEYWORD = "151"
    GENRE_KEYWORD = "155"


class GeneralMarcIDs:
    ID = ["001"]
    ID_SOURCE = ["003"]


class OrganizationMarcIDs:
    NAME = ["110", "111"]
    NAME_VARIATIONS = ["410", "411"]
    RELATED_NAMES = ["510", "511"]
    SOURCE = ["667"]
    DESCRIPTION = ["680"]
    AREA_CODE = ["043"]


class PersonMarcIDs:
    NAME = ["100"]
    NAME_VARIATIONS = ["400"]
    SOURCE = ["670"]
    DESCRIPTION = ["680"]

class TitleMarcIDs:
    NAME = ["100"]
    NAME_VARIATIONS = ["400"]
    YEAR = ["046"]
    TYPE = ["075"]

class KeywordType:
    LOC = "Kohamärksõnad"
    TIME = "Ajamärksõnad"
    TOPIC = "Teemamärksõnad"
    GENRE = "Vormimärksõnad"

class EntityType:
    PER = "PER"
    ORG = "ORG"
    KEYWORD = "EMS_KEYWORD"
    LOC = "LOC"
    UNK = "UNKNOWN"


EN_SUBJECT_FIELDS = {
    "00": "GENERAL CONCEPTS",
    "01": "PHILOSOPHY. ETHICS. SEMIOTICS",
    "02": "RELIGION. THEOLOGY. ESOTERICISM",
    "03": "SCIENCE AND TECHNOLOGY",
    "04": "EDUCATION. PEDAGOGY",
    "06": "PSYCHOLOGY",
    "07": "SOCIOLOGY. SOCIAL PSYCHOLOGY",
    "08": "EMPLOYMENT. WORKING CONDITIONS. OCCUPATIONS",
    "10": "POLITOLOGY. POLITICS",
    "11": "GOVERNMENT. ADMINISTRATION. CIVIL DEFENCE. SECURITY SERVICE",
    "12": "SOCIAL POLICY",
    "13": "LAW. LEGISLATION",
    "15": "MILITARY AFFAIRS. ARMAMENT. MILITARY EQUIPMENT",
    "17": "DEMOGRAPHY. STATISTICS",
    "18": "JOURNALISM. COMMUNICATION. MEDIA. ADVERTISING",
    "19": "NATURE PROTECTION. ENVIRONMENT PROTECTION. ENVIRONMENT TECHY",
    "20": "GEOLOGY. HYDROLOGY. CLIMATOLOGY",
    "21": "MATHEMATICS",
    "22": "ASTRONOMY. ASTROPHYSICS. SPACE EXPLORATION",
    "23": "PHYSICS",
    "24": "CHEMISTRY. CHEMICAL INDUSTRY",
    "25": "BIOLOGY. MICROBIOLOGY. GENETICS. ANTHROPOLOGY",
    "26": "BOTANY",
    "27": "ZOOLOGY",
    "28": "MEDICINE. BIOCHEMISTRY. ANATOMY. PHYSIOLOGY. COSMETICS",
    "29": "ECONOMY. ECONOMICS. FINANCE. TRADE",
    "30": "AGRICULTURE. HORTICULTURE. ANIMAL BREEDING. VETERINARY",
    "31": "FORESTRY. HUNTING. FISHERY. FOREST AND WOODWORK INDUSTRY. PULP AND PAPER INDUSTRY",
    "34": "CATERING. DOMESTIC ECONOMY",
    "35": "ELECTRICAL TECHNOLOGY. ELECTRONICS. ENERGETICS",
    "36": "INFORMATICS. INFORMATION TECHNOLOGY. AUTOMATICS",
    "39": "MECHANICAL ENGINEERING. METAL INDUSTRY. METALLURGY. MINING",
    "43": "LIGHT INDUSTRY. TEXTILE INDUSTRY. LEATHER INDUSTRY",
    "44": "BUILDING. CONSTRUCTION. SANITARY ENGINEERING",
    "45": "TRAFFIC. TRANSPORT. COMMUNICATION",
    "47": "ART. ARCHITECTURE",
    "48": "PHOTOGRAPHY. CINEMA",
    "49": "MUSIC",
    "50": "THEATRE. DANCE",
    "51": "SPORTS. PHYSICAL CULTURE",
    "52": "HOBBIES. LEISURE ACTIVITIES",
    "53": "INFORMATION SCIENCE. LIBRARIANSHIP. MEMORY INSTITUTIONS. PUBLISHING. TYPOGRAPHY",
    "54": "LINGUISTICS. LANGUAGES",
    "55": "LITERARY SCIENCE. LITERATURE. FOLKLORE",
    "56": "HISTORY. ARCHAEOLOGY",
    "58": "ETHNOLOGY. CULTURAL ANTHROPOLOGY",
    "59": "GEOGRAPHY. GEODESY. CARTOGRAPHY",
    "60": "GEOGRAPHICAL NAMES"
}

ET_SUBJECT_FIELDS = {
    "00": "ÜLDMÕISTED",
    "01": "FILOSOOFIA. EETIKA. SEMIOOTIKA",
    "02": "RELIGIOON. TEOLOOGIA. ESOTEERIKA",
    "03": "TEADUS JA TEHNIKA. TEADUSKORRALDUS. TEADUSMETODOLOOGIA. KULTUUR",
    "04": "HARIDUS. PEDAGOOGIKA",
    "06": "PSÜHHOLOOGIA",
    "07": "SOTSIOLOOGIA. SOTSIAALPSÜHHOLOOGIA",
    "08": "TÖÖHÕIVE. TÖÖTINGIMUSED. AMETID",
    "10": "RIIGIÕPETUS. POLIITIKA",
    "11": "VALITSUS. HALDUS. KODANIKUKAITSE. TURVATEENISTUS",
    "12": "SOTSIAALPOLIITIKA",
    "13": "ÕIGUS",
    "15": "SÕJANDUS. RELVAJÕUD. SÕJATEHNIKA",
    "17": "DEMOGRAAFIA. STATISTIKA",
    "18": "AJAKIRJANDUS. KOMMUNIKATSIOON. MEEDIA. REKLAAM",
    "19": "LOODUSKAITSE. KESKKONNAKAITSE. KESKKONNATEHNIKA",
    "20": "GEOLOOGIA. HÜDROLOOGIA. KLIMATOLOOGIA",
    "21": "MATEMAATIKA",
    "22": "ASTRONOOMIA. ASTROFÜÜSIKA. KOSMOSEUURIMINE",
    "23": "FÜÜSIKA",
    "24": "KEEMIA. KEEMIATÖÖSTUS",
    "25": "BIOLOOGIA. MIKROBIOLOOGIA. GENEETIKA. ANTROPOLOOGIA",
    "26": "BOTAANIKA",
    "27": "ZOOLOOGIA",
    "28": "MEDITSIIN. BIOKEEMIA. ANATOOMIA. FÜSIOLOOGIA. FARMAKOLOOGIA. KOSMEETIKA",
    "29": "MAJANDUS. MAJANDUSTEADUS. RAHANDUS. KAUBANDUS",
    "30": "PÕLLUMAJANDUS. AIANDUS. LOOMAKASVATUS. VETERINAARIA",
    "31": "METSANDUS. JAHINDUS. KALANDUS. METSA- JA PUIDUTÖÖSTUS. TSELLULOOSI- JA PABERITÖÖSTUS",
    "34": "KODUMAJANDUS. TOIDUAINETETÖÖSTUS. TOITLUSTUS. OLME",
    "35": "ELEKTROTEHNIKA. ELEKTROONIKA. ENERGEETIKA",
    "36": "INFORMAATIKA. INFOTEHNOLOOGIA. AUTOMAATIKA",
    "39": "MASINAEHITUS. METALLITÖÖSTUS. METALLURGIA. MÄENDUS",
    "43": "KERGETÖÖSTUS. TEKSTIILITÖÖSTUS. NAHA- JA JALATSITÖÖSTUS",
    "44": "EHITUS. SANITAARTEHNIKA",
    "45": "LIIKLUS. TRANSPORT. SIDE",
    "47": "KUNST. ARHITEKTUUR",
    "48": "FOTOGRAAFIA. FILM. KINO",
    "49": "MUUSIKA",
    "50": "TEATER. TANTS",
    "51": "SPORT. KEHAKULTUUR",
    "52": "HARRASTUSED. VABA AEG",
    "53": "INFOTEADUS. RAAMATUKOGUNDUS. MÄLUASUTUSED. KIRJASTAMINE. TRÜKINDUS",
    "54": "KEELETEADUS. KEELED",
    "55": "KIRJANDUSTEADUS. ILUKIRJANDUS. RAHVALUULE",
    "56": "AJALUGU. ARHEOLOOGIA",
    "58": "ETNOLOOGIA. KULTUURIANTROPOLOOGIA",
    "59": "GEOGRAAFIA. GEODEESIA. KARTOGRAAFIA",
    "60": "KOHANIMED"
}



from dataclasses import dataclass

COMPONENT_KEY = "meta_extractor"


class Tasks:
    SINGLE = "extract_meta_from_text"
    PIPELINE = "run_meta_extractor_with_core_logic"


class Queue:
    MAIN = "meta_extractor"


class StatusKeys:
    EXTRACT_METADATA = "extract_metadata"

class Error:
    UNKNOWN = "Failed to extract meta information from digitizer output!"


@dataclass(frozen=True)
class TitleType:
    AUTHOR_WITHOUT_TITLE: str = "pealkirjata autor"
    NORMALIZED_TITLE: str = "normitud eelispealkiri"
    TITLE: str = "väljaandes esitatud kujul põhipealkiri"
    PARALLEL_TITLE: str = "rööppealkiri"
    ADDITIONAL_TITLE: str = "alampealkiri"
    METS_TITLE: str = "väljaandes esitatud kujul põhipealkiri"
    ANON: str = "anonüümne väljaanne"


TITLE_TYPES_MAP = {
    TitleType.AUTHOR_WITHOUT_TITLE: 130,
    TitleType.NORMALIZED_TITLE: 240,
    TitleType.TITLE: 245,
    TitleType.PARALLEL_TITLE: 246,
    TitleType.ADDITIONAL_TITLE: 245,
    TitleType.METS_TITLE: 245,
    TitleType.ANON: 130
}


PUBLISHER_KEY = "Väljaandja"



from rara_tools.exceptions import SierraResponseConverterException


class SierraResponseConverter:
    """Converts a JSON response from the Sierra API to MARC-in-JSON format."""

    def __init__(self, response: dict):
        if not isinstance(response, dict):
            raise SierraResponseConverterException(
                "Please provide a valid JSON response.")
        self.response = response

    @staticmethod
    def _map_control_fields(field: dict) -> dict:
        # for tags < 010, no subfields, instead one str value in "value"
        return {field["tag"]: field["value"]}
    
    def _map_data_field_subfields(self, subfields: list) -> list:
        """ Maps data field subfields to the standardised format, if necessary.

        Args:
            subfields (list): List of subfields in format e.g
              "subfields": [
                        {
                            "code": "a",
                            "data": "foo"
                        }, 
                        {
                            "code": "c",
                            "data": "by me."
                        }
                    ]

        Returns:
            list: standardised marc-in-json format. e.g 
             "subfields": [
                            { "a": "foo /" },
                            { "c": "by me." }
                                ]
        """
        result = []
        for subfield in subfields:
            
            code = subfield.get("code")
            data = subfield.get("data")

            if not code or not data:
                # assume that the subfield is already in the correct format
                return subfields
            else:
                result.append({code: data})
                
        return result


    def _map_data_fields(self, field: dict) -> dict:
        """ Maps marc fields > 010. 

        Args:
            field (dict): Contains the marc tag and list with indicators and subfields.

        Returns:
            dict: standardised marc-in-json format.
        """

        data = field["data"]
        subfields = data.get("subfields", [])

        # Order matters ind1, in2, subfields
        field_data = {
            "ind1": data.get("ind1", " "),
            "ind2": data.get("ind2", " "),
            "subfields": self._map_data_field_subfields(subfields)
        }

        return {field["tag"]: field_data}

    @staticmethod
    def _is_marc21structured(field: dict) -> bool:
        """Checks if the field is already structured according to MARC21 in JSON"""
        return any(key.isdigit() for key in field.keys())
    
    def _handle_field_type(self, field: dict) -> dict:

        if self._is_marc21structured(field):
            return field

        if field.get("data"):
            return self._map_data_fields(field)

        tag = field.get("tag")

        if not tag:
            raise SierraResponseConverterException(
                "Field is missing MARC21 tag.")

        if tag < "010":
            return self._map_control_fields(field)
        else:
            return self._map_data_fields(field)

    def _convert_response(self) -> list:
        entries = self.response.get("entries")
        if not entries:
            raise SierraResponseConverterException(
                "No entries found in the response.")

        try:
            return [
                {
                    "sierraID": str(e["id"]),
                    "leader": e["marc"]["leader"],
                    "fields": [
                        self._handle_field_type(f) for f in e["marc"]["fields"]
                    ]}
                for e in entries
            ]

        except KeyError as e:
            raise SierraResponseConverterException(
                f"Malformed response: missing key {e}")

    def convert(self) -> list:
        try:
            return self._convert_response()
        except Exception as e:
            raise SierraResponseConverterException(
                f"An unexpected error occurred: {e}")



from typing import List, NoReturn
from pymarc.record import Record
from rara_tools.parsers.tools.entity_normalizers import KeywordNormalizer
from rara_tools.parsers.marc_records.base_record import BaseRecord
from rara_tools.constants.parsers import (
    EMSMarcIDs, KeywordType,
    EN_SUBJECT_FIELDS, ET_SUBJECT_FIELDS
)
import regex as re
import json



class EMSRecord(BaseRecord):
    """ Generates a simplified EMS JSON record
    from a pymarc MARC record.
    """
    def __init__(self, record: Record, add_variations: bool = False) -> NoReturn:
        """ Initializes EMSRecord object.

        Parameters
        -----------
        record: Record
            pymarc.record.Record object.
        add_variations: bool
            If enabled, constructs an additional variations field, which
            combines the content of multiple fields + adds some generated
            variations. If the output is uploaded into Elastic and used
            via rara-norm-linker, it is necessary to enable this.
        """
        super().__init__(record=record, add_variations=add_variations)
        self.__en_subject_field_mapping: dict = EN_SUBJECT_FIELDS
        self.__et_subject_field_mapping: dict = ET_SUBJECT_FIELDS
        self.__keyword: str = ""
        self.__keyword_en: str = ""
        self.__keyword_type: str = ""
        self.__keyword_variations: List[str] = []
        self.__keyword_fields = {
            EMSMarcIDs.TIME_KEYWORD: KeywordType.TIME,
            EMSMarcIDs.TOPIC_KEYWORD: KeywordType.TOPIC,
            EMSMarcIDs.LOC_KEYWORD: KeywordType.LOC,
            EMSMarcIDs.GENRE_KEYWORD: KeywordType.GENRE
        }

        self.__ems_url_ids = EMSMarcIDs.URL
        self.__synonym_ids = EMSMarcIDs.SYNONYMS
        self.__related_ids = EMSMarcIDs.RELATED
        self.__category_ids = EMSMarcIDs.CATEGORY
        self.__notes_ids = EMSMarcIDs.NOTES
        self.__synonyms: List[str] = []
        self.__synonyms_en: List[str] = []
        self.__subject_field_ids: List[str] = []

        self.__subject_fields_et: List[str] = []
        self.__subject_fields_en: List[str] = []
        self.__ems_url: str = ""
        self.__narrower: List[str] = []
        self.__broader: List[str] = []
        self.__related: List[str] = []
        self.__narrower_ems_urls: List[str] = []
        self.__broader_ems_urls: List[str] = []
        self.__related_ems_urls: List[str] = []
        self.__variations: List[str] = []
        self.__variations_en: List[str] = []
        self.__use_with_others: bool | None = None
        self.__full_record: dict = {}

    @property
    def keyword(self) -> str:
        if not self.__keyword:
            self.__keyword = self.get_values(
                marc_ids=self.__keyword_fields,
                subfield_id="a"
            )[0]
        return self.__keyword

    @property
    def keyword_en(self) -> str:
        if not self.__keyword_en:
            self.__keyword_en = self.synonyms_en[0] if self.synonyms_en else ""
        return self.__keyword_en

    @property
    def synonyms(self) -> List[str]:
        if not self.__synonyms:
            self.__synonyms = self.get_values(
                marc_ids=self.__synonym_ids,
                subfield_id="a"
            )
        return self.__synonyms

    @property
    def synonyms_en(self) -> List[str]:
        if not self.__synonyms_en:
            self.__synonyms_en = self.get_values(
                marc_ids=self.__synonym_ids,
                subfield_id="a", ind2="9"
            )
        return self.__synonyms_en

    @property
    def subject_field_ids(self) -> List[str]:
        if not self.__subject_field_ids:
            self.__subject_field_ids = self.get_values(
                marc_ids=self.__category_ids,
                subfield_id="a",
                ind2="7"
            )
        return self.__subject_field_ids

    @property
    def subject_fields_et(self) -> List[str]:
        if not self.__subject_fields_et:
            self.__subject_fields_et = [
                self.__et_subject_field_mapping[_id]
                for _id in self.subject_field_ids
            ]
        return self.__subject_fields_et

    @property
    def subject_fields_en(self) -> List[str]:
        if not self.__subject_fields_en:
            self.__subject_fields_en = [
                self.__en_subject_field_mapping[_id]
                for _id in self.subject_field_ids
            ]
        return self.__subject_fields_en


    @property
    def ems_url(self) -> str:
        if not self.__ems_url:
            self.__ems_url = self.get_values(
                marc_ids=self.__ems_url_ids,
                subfield_id="0",
                ind2="8"
            )[0]
        return self.__ems_url

    @property
    def broader(self) -> List[str]:
        if not self.__broader:
            self.__broader = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="a",
                subfield_restriction = ("w", "g")
            )
        return self.__broader

    @property
    def narrower(self) -> List[str]:
        if not self.__narrower:
            self.__narrower = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="a",
                subfield_restriction = ("w", "h")
            )
        return self.__narrower

    @property
    def related(self) -> List[str]:
        if not self.__related:
            self.__related = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="a",
                subfield_to_ignore ="w"
            )
        return self.__related

    @property
    def broader_ems_urls(self) -> List[str]:
        if not self.__broader_ems_urls:
            self.__broader_ems_urls = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="0",
                subfield_restriction = ("w", "g")
            )
        return self.__broader_ems_urls

    @property
    def narrower_ems_urls(self) -> List[str]:
        if not self.__narrower_ems_urls:
            self.__narrower_ems_urls = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="0",
                subfield_restriction = ("w", "h")
            )
        return self.__narrower_ems_urls

    @property
    def related_ems_urls(self) -> List[str]:
        if not self.__related_ems_urls:
            self.__related_ems_urls = self.get_values(
                marc_ids=self.__related_ids,
                subfield_id="0",
                subfield_to_ignore ="w"
            )
        return self.__related_ems_urls

    @property
    def keyword_type(self) -> str:
        if not self.__keyword_type:
            for field in self.dict_record:
                field_id = list(field.keys())[0]
                if field_id in self.__keyword_fields:
                    self.__keyword_type = self.__keyword_fields[field_id]
        return self.__keyword_type

    @property
    def use_with_others(self) -> bool:
        if self.__use_with_others == None:
            notes = self.get_values(marc_ids=self.__notes_ids, subfield_id="i")
            self.__use_with_others = False
            if notes:
                if re.search(r"Kasutada koos teise", notes[0]):
                    self.__use_with_others = True

        return self.__use_with_others

    @property
    def variations(self) -> List[str]:
        if not self.__variations:
            original_variations = self.synonyms + [self.keyword]
            variations = []
            for kw in original_variations:
                variations_ = KeywordNormalizer(kw, keyword_type=self.keyword_type).variations
                variations.extend(variations_)
            self.__variations = [v.lower() for v in list(set(variations))]
        return self.__variations

    @property
    def variations_en(self) -> List[str]:
        if not self.__variations_en:
            pass
        return self.__variations_en


    @property
    def full_record(self) -> dict:
        if not self.__full_record:
            self.__full_record = {
                "keyword": self.keyword,
                "keyword_en": self.keyword_en,
                "keyword_type": self.keyword_type,
                "use_with_others": self.use_with_others,
                "subject_field_ids": self.subject_field_ids,
                "subject_fields_et": self.subject_fields_et,
                "subject_fields_en": self.subject_fields_en,
                "synonyms": self.synonyms,
                "synonyms_en": self.synonyms_en,
                "narrower": self.narrower,
                "broader": self.broader,
                "related": self.related,
                "narrower_ems_urls": self.narrower_ems_urls,
                "broader_ems_urls": self.broader_ems_urls,
                "related_ems_urls": self.related_ems_urls,
                "ems_id": self.identifier,
                "ems_url": self.ems_url,
                "identifier_source": self.identifier_source,
                "full_record_marc": str(self.marc_record),
                "full_record_json": json.dumps(self.marc_json_record)
            }
            if self.add_variations:
                self.__full_record.update(
                    {"link_variations": self.variations}
                )
        return self.__full_record



import os
import json

from rara_tools.constants import YYMMDD_FORMAT
from rara_tools.normalizers import BibRecordNormalizer, AuthoritiesRecordNormalizer
from tests.test_utils import (
    get_linker_res_example,
    get_formatted_sierra_response,
    check_record_tags_sorted,
    check_no_dupe_tag_values,
    check_record_tags_have_values,
)
from rara_tools.normalizers.viaf import VIAFRecord

from rara_tools.constants.linker import EntityType

from datetime import datetime

from pymarc import Record, JSONReader


TEST_LEVEL = os.getenv("TEST_LEVEL", "unit")

EMPTY_SIERRA_RECORDS = [
    {"sierraID": "1", "leader": "00000nz  a2200000n  4500", "fields": []},
]

REQUIRED_FIELDS = ["667", "925"]  # always included after normalization
MOCK_LINKER_ONE_FOUND = get_linker_res_example("oneFound.json")
MOCK_LINKER_MULTIPLE_FOUND = get_linker_res_example("multipleFound.json")
MOCK_LINKER_NOT_FOUND = get_linker_res_example("notFound.json")

def get_100_field(author: str = "Autor", title: str = "Pealkiri") -> dict:
        return {
            "100": {
                "ind1": "1",
                "ind2": " ",
                "subfields": [{"a": author, "t": title}],
            }
        }


def test_normalizers_OK():
    """Test field editing logic & internals"""

    linking_results = [MOCK_LINKER_ONE_FOUND, MOCK_LINKER_MULTIPLE_FOUND]

    test_sierra_data = get_formatted_sierra_response("authorities.json")

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )

    assert len(normalizer.records_extra_data) == len(normalizer.data)

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
        sierra_data=test_sierra_data,
    )
    assert len(normalizer.records_extra_data) == len(normalizer.data)

    data = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {"667": {"ind1": " ", "ind2": " ", "subfields": [{"a": "Val"}]}},
            ],
        },
    ]

    # default behavior - added if not in record &
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data,
        ALLOW_EDIT_FIELDS=[],
        REPEATABLE_FIELDS=[],
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # not edited if exists
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data, ALLOW_EDIT_FIELDS=[], REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        assert r.get_fields("667")[0].get_subfields("a")[0] == "Val"

    # allow repeatable, new field will be added
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data, ALLOW_EDIT_FIELDS=[], REPEATABLE_FIELDS=["667"]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 2
        assert fields_667[0].get_subfields("a")[0] == "Val"
        assert fields_667[1].get_subfields("a")[0] == "Muudetud AI poolt"

    # allow editing, field will be edited
    normalizer = AuthoritiesRecordNormalizer(
        sierra_data=data, ALLOW_EDIT_FIELDS=["667"], REPEATABLE_FIELDS=[]
    )
    for r in normalizer:
        fields_667 = r.get_fields("667")
        assert len(fields_667) == 1
        assert fields_667[0].get_subfields("a")[0] == "Muudetud AI poolt"


def _validate_existing_record_notes(record: Record):
    """Validate notes in existing record - 667 & 925 fields"""

    fields_667 = record.get_fields("667")
    assert any("Muudetud AI poolt" in f.get_subfields("a") for f in fields_667)

    fields_925 = record.get_fields("925")
    assert len(fields_925) == 1
    assert len(fields_925[0].get_subfields("p")) == 1
    assert len(fields_925[0].get_subfields("t")) == 0


def _validate_new_record_notes(record: Record):
    """Validate notes in new record - 667 & 925 fields"""

    fields_667 = record.get_fields("667")

    assert any("Loodud AI poolt" in f.get_subfields("a") for f in fields_667)

    fields_925 = record.get_fields("925")
    assert len(fields_925) == 1
    assert len(fields_925[0].get_subfields("p")) == 0
    assert len(fields_925[0].get_subfields("t")) == 1


def validate_bibrecord_normalized(record: Record, has_viaf_data=False):
    """Validate specific fields in normalized bibrecord"""

    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40
    # pos 00-05 is current date in YYMMDD format
    timestamp = field_008[0:6]
    # check timestamp in correct format
    try:
        datetime.strptime(timestamp, YYMMDD_FORMAT)
    except ValueError:
        raise AssertionError(
            f"008 field timestamp {timestamp} is not in format {YYMMDD_FORMAT}"
        )

    expected_suffix_008 = "|||aznnnaabn          || |||" + 6 * " "
    suffix = field_008[6:]
    assert suffix == expected_suffix_008

    _validate_existing_record_notes(record)


def validate_authorities_record_normalized(record: Record, has_viaf_data=False):
    """Validate specific fields in normalized authority record"""

    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40
    # pos 00-05 is current date in YYMMDD format
    timestamp = field_008[0:6]
    # check timestamp in correct format
    try:
        datetime.strptime(timestamp, YYMMDD_FORMAT)
    except ValueError:
        raise AssertionError(
            f"008 field timestamp {timestamp} is not in format {YYMMDD_FORMAT}"
        )

    expected_suffix_008 = "|n|adnnnaabn          || |a|" + 6 * " "
    suffix = field_008[6:]
    assert suffix == expected_suffix_008

    _validate_existing_record_notes(record)

    # check that a, b & c subfields have values (can have default or unique)
    field_040_subfields = record.get_fields("040")[0]
    assert len(field_040_subfields.get_subfields("a")) > 0
    assert len(field_040_subfields.get_subfields("b")) > 0
    assert len(field_040_subfields.get_subfields("c")) > 0

    # check that 008 field has a value of length 40
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40

    if has_viaf_data:
        field_043 = record.get_fields("043")[0].get_subfields("c")[
            0
        ]  # check that 043 has subfield c with value "ee"
        assert field_043 == "ee"

        field_024 = record.get_fields("024")
        for f in field_024:
            assert len(f.get_subfields("0")) > 0  # VIAF url

        field_046 = record.get_fields("046")[0]
        assert len(field_046.get_subfields("f")) > 0  # birth date
        assert len(field_046.get_subfields("g")) > 0  # death date


def test_missing_fields_created_bibrecord_normalization():
    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = BibRecordNormalizer(
        linking_results=linking_results,
    )

    normalizer_sierra_data_only = BibRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for record in normalizer_entities_only:
        check_record_tags_have_values(
            record,
            [
                "008",  # Sierra related, always with bibs
                "100",  # VIAf enriched
            ]
            + REQUIRED_FIELDS,
        )
        validate_bibrecord_normalized(record, has_viaf_data=True)

    for record in normalizer_sierra_data_only:
        check_record_tags_have_values(
            record,
            [
                "008"  # Sierra related, always with bibs
            ]
            + REQUIRED_FIELDS,
        )
        validate_bibrecord_normalized(record)


def test_missing_fields_created_authorities_normalization():
    linking_results = [MOCK_LINKER_ONE_FOUND]

    normalizer_entities_only = AuthoritiesRecordNormalizer(
        linking_results=linking_results,  # find one match
    )

    normalizer_sierra_data_only = AuthoritiesRecordNormalizer(
        sierra_data=EMPTY_SIERRA_RECORDS,
    )

    for r in normalizer_entities_only:
        check_record_tags_have_values(
            r,
            [
                "008",
                "040",  # SIERRA related
                "024",
                "043",
                "046",  # VIAF enriched
            ]
            + REQUIRED_FIELDS,
        )

        validate_authorities_record_normalized(r, True)

    for r in normalizer_sierra_data_only:
        check_record_tags_have_values(r, ["040"] + REQUIRED_FIELDS)
        validate_authorities_record_normalized(r)


def test_normalized_fields_sorted():
    unsorted_bibdata = [
        {
            "sierraID": "1",
            "leader": "00000nz  a2200000n  4500",
            "fields": [
                {
                    "035": {
                        "ind1": " ",
                        "ind2": " ",
                        "subfields": [{"a": "(ErESTER)<1>"}],
                    }
                },
                {"008": "220805|||aznnnaabn          || |||      nz n  "},
                {"046": {"ind1": " ", "ind2": " ", "subfields": [{"k": "1912"}]}},
            ],
        }
    ]

    normalizers = (BibRecordNormalizer, AuthoritiesRecordNormalizer)

    for normalizer in normalizers:
        normalizer = normalizer(linking_results=[], sierra_data=unsorted_bibdata)

        for r in normalizer:
            check_no_dupe_tag_values(r)
            check_record_tags_sorted(r)


def test_authority_normrecord_found_in_es_and_normalized():
    """KATA elastic normkirjete seast leitakse 1 vaste & normaliseerija täiendab leitud normkirjet VIAF infoga.
    - valideeri normaliseerimise mapping, mis autori tabelis. Täiendatud väljad ja VIAFist info"""
    # Presume, author name identified and sent to linker
    linker_res = get_linker_res_example("oneFound.json")  # single result
    linking_results = [linker_res]

    # 1 result found
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)

    data = normalizer.data

    assert len(data) == 1
    record = next(iter(JSONReader(json.dumps(normalizer.data, ensure_ascii=False))))
    check_record_tags_have_values(record, ["040"] + REQUIRED_FIELDS)
    validate_authorities_record_normalized(record, has_viaf_data=True)


def _run_normalizer(linked_data):
    normalizer = AuthoritiesRecordNormalizer(linking_results=linked_data)
    return normalizer.data


def test_normalizer_handles_bad_inputs():
    linker_res = get_linker_res_example("oneFound.json")

    # pop the leader field to simulate record without leader
    linker_res["linked_info"][0]["json"].pop("leader", None)
    _run_normalizer([linker_res])

    # make fields empty to simulate a record with no fields
    linker_res["linked_info"][0]["json"]["fields"] = []
    _run_normalizer([linker_res])

    # pop the fields to simulate a record with no fields
    linker_res["linked_info"][0]["json"].pop("fields", None)
    _run_normalizer([linker_res])

    inputs = ["", None, [], {}, 123]

    _run_normalizer(inputs)


def test_matching_sierra_record_viaf_id_found():
    """normkirjelt leitakse VIAF ID, vajadusel normi asukoht, kus see ID sisaldub."""
    pass


def test_matching_sierra_record_viaf_id_not_found():
    """kirjelt VIAF IDd ei leitud, soorita otsing VIAFi pihta, et leida _vastutav isik_?. Loo uus vastavalt otsingu tulemusele."""
    pass


def test_authorities_normalizer_checks():
    """
    - kontrolli kas tuvastatud nimi on SIERRAst leitud vaste 1XX, 4XX väljadel. Kui pole, siis lisa 4XX väljale.
    - kontrolli, kas VIAF andmete nimekujud on normkandes olemas. Kui pole, lisa need 4XX väljale.
    - Kontrolli, kas VIAF kandes on sünni ja surma daatumid ja kas need klapivad normkandes olevaga. Kui pole, siis liiguta normkandest kogu 1XX väli 4XX väljale. Seejärel loo uute daatumitega 1XX väli.
    - Kontrolli, et väljal 046 olevad daatumid klapiksid just 1xx väljale lisatuga. Kui andmeid muudeti, siis märgi, et baasis on normkanne muutunud
    """
    pass


def test_add_birth_and_death_dates():
    """Test adding birth & death dates to 046 field from VIAF record. Expected Date format YYMMDD"""

    # Case one: viaf record has both birth & death date
    normalizer = AuthoritiesRecordNormalizer()
    record = Record()

    viaf_record = normalizer._get_viaf_record(
        record,
        False,
        # entity="Paul Keres"
        viaf_id=7432247,
    )
    for value in [viaf_record, viaf_record.birth_date, viaf_record.death_date]:
        assert value is not None

    normalizer._add_birth_and_death_dates(record, viaf_record)

    field_046 = str(record.get_fields("046")[0])
    assert field_046 == "=046  \\\\$f19160107$g19750605"

    # Case two: viaf record has birth date, but no death date (author still alive)
    viaf_record = normalizer._get_viaf_record(
        record,
        False,
        #   entity="Andrus Kivirähk"
        viaf_id=38111668,
    )

    assert viaf_record is not None
    assert viaf_record.birth_date is not None
    assert viaf_record.death_date == 0

    record = Record()
    normalizer._add_birth_and_death_dates(record, viaf_record)

    field_046 = str(record.get_fields("046")[0])
    # empty indicators represented with \
    assert field_046 == "=046  \\\\$f19700817"

    # Case 3 - viaf record has no birth or death date
    viaf_record = normalizer._get_viaf_record(
        False,
        record,
        viaf_id=1165154688467248380004,
        # entity="Eesti Interlingvistika Selts"
    )
    record = Record()
    normalizer._add_birth_and_death_dates(record, viaf_record)
    # should not add 046 field
    fields_046 = record.get_fields("046")
    assert len(fields_046) == 0


def test_add_nationality():
    """Test adding nationality from VIAF record to 043 field"""

    # Case 1: nationality is not estonian - do not add 043 field

    normalizer = AuthoritiesRecordNormalizer()
    record = Record()

    viaf_record = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=52700,
        #   entity="Paulo Coelho"
    )

    assert viaf_record is not None
    assert viaf_record.nationality == "br"

    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 0

    # Case 2, nationality missing

    normalizer = AuthoritiesRecordNormalizer()
    record = Record()

    viaf_record = normalizer._get_viaf_record(
        record, False, viaf_identifier="123zsxcasd"
    )

    assert viaf_record is None
    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 0

    # Case 3 - Nationality is Estonian - add 043 field
    viaf_record = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=249285257,
        #   entity="Jaan Tõnisson"
    )

    assert viaf_record is not None
    assert viaf_record.nationality == "ee"

    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 1
    assert str(fields_043[0]) == "=043  \\\\$cee"

    # Case 4 - 043 field already exists - should not get edited (not in ALLOW_EDIT_FIELDS)
    linking_results = [
        {
            "original_entity": "Eduard Vilde",
            "entity_type": EntityType.PER,
            "linked_info": [
                {
                    "json": {
                        "leader": "00736nz  a2200217n  4500",
                        "fields": [
                            {
                                "043": {
                                    "ind1": " ",
                                    "ind2": " ",
                                    "subfields": [{"c": "gb"}],
                                }
                            }
                        ],
                    }
                }
            ],
        }
    ]
    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
    )
    record = normalizer.first
    # mock run add nationality with foreign VIAF record
    viaf_record = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=249285257,
        #   entity="Jaan Tõnisson"
    )
    normalizer._add_nationality(record, viaf_record)
    fields_043 = record.get_fields("043")
    assert len(fields_043) == 1
    assert str(fields_043[0]) == "=043  \\\\$cgb"


def test_create_new_normrecord():
    # Case 1 - one entity found - update existing norm record
    linker_res = get_linker_res_example("oneFound.json")
    linking_results = [linker_res]

    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data

    assert len(data) == 1  # should update existing record

    # Case 2 - entities not found - create new norm record
    linking_results = [MOCK_LINKER_NOT_FOUND]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1

    record = normalizer.first
    leader = str(record.leader)
    assert leader == "01682nz  a2200349n  4500"
    assert len(leader) == 24
    _validate_new_record_notes(record)

    # Case 3 - multiple entities found - record creation will be skipped
    linker_res = get_linker_res_example("multipleFound.json")
    linking_results = [linker_res]
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 0

    # Case 3 with bibnormalizer - update with better example
    linking_results = [MOCK_LINKER_NOT_FOUND]
    
    classified_fields = [
        [get_100_field()]
    ]

    normalizer = BibRecordNormalizer(linking_results=linking_results, classified_fields=classified_fields)
    data = normalizer.data
    assert len(data) == 1
    record = normalizer.first

    _validate_new_record_notes(record)
    # validate leader
    leader = str(record.leader)
    assert leader == "00399nz  a2200145n  4500"
    assert len(leader) == 24

    # Mock not found title
    linking_results = [
        {
            "original_entity": "Eesti Ekspress",
            "entity_type": EntityType.TITLE,
            "linked_info": [],
        }
    ]
    normalizer = BibRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    assert len(data) == 1
    record = normalizer.first

    # Test 100|d gets date
    linking_results = [
        {
            "original_entity": "Libe, Katariina",
            "entity_type": EntityType.PER,
            "linked_info": [],
        }
    ]

    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    data = normalizer.data
    record = normalizer.first
    # Check that 100|d has date added from VIAF
    field_100 = record.get_fields("100")[0]
    assert field_100.get_subfields("d")[0] == "1986-"
    field_046 = record.get_fields("046")[0]
    assert field_046.get_subfields("f")[0] == "19861126"


def test_680_field_on_existing_record_moved_to_667():
    """680 Should not be added for new, if exists on existing record, should be moved to 667"""
    linker_res = get_linker_res_example("oneFound.json")

    # add 680 field to existing record will have two 680 fields after adding
    linker_res["linked_info"][0]["json"]["fields"].append(
        {"680": {"ind1": " ", "ind2": " ", "subfields": [{"a": "Test 680 field"}]}}
    )
    initial_record = JSONReader(
        json.dumps([linker_res["linked_info"][0]["json"]], ensure_ascii=False)
    )
    record = next(iter(initial_record))
    fields_680 = record.get_fields("680")
    assert len(fields_680) == 2  # inital plus new
    fields_667 = record.get_fields("667")
    assert len(fields_667) == 0  # initial only

    linking_results = [linker_res]

    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    record = normalizer.first

    fields_680 = record.get_fields("680")
    assert len(fields_680) == 0
    fields_667 = record.get_fields("667")
    assert len(fields_667) == 3  # original + moved from 680 + new note


def test_date_formatting():
    normalizer = AuthoritiesRecordNormalizer()

    dates = {
        "19700712": "19700712",
        "1970": "1970",
        "1970-07": "197007",
        "2001-12-31": "20011231",
        "1999-01": "199901",
    }

    for input_date, expected in dates.items():
        assert normalizer._format_date(input_date) == expected

    # invalid date formats - should return empty string
    invalid_dates = ["abcd", "199A0101"]
    for date in invalid_dates:
        assert normalizer._format_date(date) == ""


def test_new_bibrecord_title_included():
    """normrecord for bibs has to always have the 1XX|t field filled"""

    # Case 1 No linker response, & Viaf record found
    linking_results = [
        {
            "original_entity": "Lord of the Rings",
            "entity_type": EntityType.TITLE,
            "linked_info": [],
        }
    ]
    # classified_fields = [
    #     [get_100_field("Lord of the Rings")]
    # ]

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
        # classified_fields=classified_fields,
    )
    data = normalizer.data
    assert len(data) == 1  # should enrich existing record
    record = normalizer.first

    _validate_new_record_notes(record)
    fields_100 = record.get_fields("100")
    assert len(fields_100) == 1
    assert fields_100[0].get_subfields("t")[0] == "Lord of the rings"

    # Case 2 - Viaf record not found - should use classified author
    linking_results = [
        {
            # "original_entity": "Roolijoodiku katastroofiline jõulusõit",
            "entity_type": EntityType.TITLE,
            "linked_info": [],
        }
    ]
    classified_fields = [
        [get_100_field(title="Roolijoodiku katastroofiline jõulusõit")]
    ]
    
    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
        classified_fields=classified_fields,
    )
    record = normalizer.first
    data = normalizer.data
    assert len(data) == 1  # should enrich existing record

    fields_100 = record.get_fields("100")
    assert len(fields_100) == 1
    assert (
        fields_100[0].get_subfields("t")[0] == "Roolijoodiku katastroofiline jõulusõit"
    )
    
    # withourt classified fields - should not create record
    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
    )
    data = normalizer.data
    assert len(data) == 0  # should not create record


def _validate_new_record_008_field(record: Record):
    """Validate 008 field in new record"""
    field_008 = record.get_fields("008")[0].data
    assert len(field_008) == 40
    # pos 00-05 is current date in YYMMDD format
    timestamp = field_008[0:6]
    try:
        datetime.strptime(timestamp, YYMMDD_FORMAT)
    except ValueError:
        raise AssertionError(
            f"008 field timestamp {timestamp} is not in format {YYMMDD_FORMAT}"
        )


def test_008_field_formatting():
    """00-04 position will be changed for new record, not edited on existing record"""

    # Case 1 - new record created, should have current date in 008 field

    linking_results = [
        {
            "original_entity": "Eesti Ekspress",
            "entity_type": EntityType.TITLE,
            "linked_info": [],
        }
    ]

    normalizer = BibRecordNormalizer(
        linking_results=linking_results,
    )
    new_record = normalizer.first
    _validate_new_record_008_field(new_record)

    # Case 2 - existing record updated, 008 field should not be changed
    linker_res = get_linker_res_example("oneFound.json")
    linking_results = [linker_res]
    original_record = JSONReader(
        json.dumps([linker_res["linked_info"][0]["json"]], ensure_ascii=False)
    )
    record = next(iter(original_record))
    original_008 = record.get_fields("008")[0].data

    # for authorities
    normalizer = AuthoritiesRecordNormalizer(linking_results=linking_results)
    authorities_record = normalizer.first
    field_008 = authorities_record.get_fields("008")[0].data
    assert len(field_008) == 40

    assert field_008 == original_008

    # for bibs
    normalizer = BibRecordNormalizer(linking_results=linking_results)
    expected_008 = "990107|||aznnnaabn          || |||" + 6 * " "
    bibrecord = normalizer.first
    field_008 = bibrecord.get_fields("008")[0].data
    assert len(field_008) == 40
    assert field_008 == expected_008


def test_classified_fields_added_to_linked_record():
    """Test that classified fields Can be passed to normalizer & added to linked record"""

    classified_fields = (
        [
            {
                "046": {"ind1": " ", "ind2": " ", "subfields": [{"k": "2021"}]},
                "670": {
                    "ind1": " ",
                    "ind2": "0",
                    "subfields": [{"a": "Päikesekiri, 2021"}],
                },
                "100": {
                    "ind1": "2",
                    "ind2": " ",
                    "subfields": [{"a": "Eesti Kirjandusmuuseum"}],
                },
            }
        ],
    )
    # Case 1 - no 670 exists, should be added to linked record
    for normalizer in (AuthoritiesRecordNormalizer, BibRecordNormalizer):
        linking_results = [MOCK_LINKER_NOT_FOUND]
        normalizer = normalizer(
            linking_results=linking_results, classified_fields=classified_fields
        )

        record = normalizer.first
        fields_670 = record.get_fields("670")
        assert len(fields_670) == 1
        assert fields_670[0].get_subfields("a")[0] == "Päikesekiri, 2021"

    # Case 2 - existing record with 670 should not update (same behavior for both normalizers)
    linker_res = get_linker_res_example("oneFound.json")
    linking_results = [linker_res]

    for normalizer in (AuthoritiesRecordNormalizer,
                       BibRecordNormalizer):
        normalizer = normalizer(
            linking_results=linking_results, classified_fields=classified_fields
        )
        record = normalizer.first
        fields_670 = record.get_fields("670")
        assert len(fields_670) == 1
        assert fields_670[0].get_subfields("a")[0] == "Eesti kirjarahva leksikon, 1995."

    def get_046_field(year: str) -> dict:
        return {"046": {"ind1": " ", "ind2": " ", "subfields": [{"k": year}]}}
    
    # Case 3 - 046 $k - publication date Passed for bib
    classified_fields = [
        [get_046_field("2021"), get_100_field("Title 2021")],
        [get_046_field("1999"), get_100_field("Title 1999")],
        [get_046_field("2022"), get_100_field("Title 2022")],
    ]

    mock_046_exists = MOCK_LINKER_ONE_FOUND.copy()
    mock_046_exists["linked_info"][0]["json"]["fields"].append(get_046_field("2000"))

    # for new record should get included
    linking_results = [
        MOCK_LINKER_NOT_FOUND,  # new record
        MOCK_LINKER_ONE_FOUND,  # new record
        MOCK_LINKER_NOT_FOUND,
    ]  # editing existing record

    normalizer = BibRecordNormalizer(
        linking_results=linking_results, classified_fields=classified_fields
    )

    # for i, record in enumerate(normalizer):
    # first two should have 046 from classified data
    record1 = normalizer.get_record(0)
    fields_046 = record1.get_fields("046")
    assert len(fields_046) == 1
    assert fields_046[0].get_subfields("k")[0] == "2021"

    record2 = normalizer.get_record(1)
    fields_046 = record2.get_fields("046")
    assert len(fields_046) == 1
    # should be unchanged, aka 2000
    assert fields_046[0].get_subfields("k")[0] == "2000"

    record3 = normalizer.get_record(2)
    fields_046 = record3.get_fields("046")
    assert len(fields_046) == 1
    assert fields_046[0].get_subfields("k")[0] == "2022"


def test_classified_author_included_to_new_bibrecord():
    # Case 1 - new record created, should have 100|a from classified data
    expected_author = "Primary Author"
    expected_date = "2002-"
    expected_title = "Title"
    expected_role = "autor"
    classified_fields = [
        [
            {
                "100": {
                    "ind1": "2",
                    "ind2": " ",
                    "subfields": [
                        {"a": expected_author},
                        {"d": expected_date},
                        {"e": expected_role},
                        {"t": expected_title},
                    ],
                }
            },
        ],
        [
            {
                "670": {
                    "ind1": " ",
                    "ind2": " ",
                    "subfields": [{"a": "Another Task, 2022"}],
                }
            },
            {"046": {"ind1": " ", "ind2": " ", "subfields": [{"k": "2022"}]}},
            {
                "100": {
                    "ind1": "2",
                    "ind2": " ",
                    "subfields": [
                        {"a": expected_author},
                        {"d": expected_date},
                        {"e": expected_role},
                        {"t": expected_title},
                    ],
                }
            },
        ],
        [],
    ]

    linking_results = [
        MOCK_LINKER_NOT_FOUND,  # create new
        MOCK_LINKER_ONE_FOUND,  # update existing
        MOCK_LINKER_NOT_FOUND,  # create new
    ]

    normalizer = BibRecordNormalizer(
        linking_results=linking_results, classified_fields=classified_fields
    )

    data = normalizer.data
    assert len(data) == 2  # two records created/updated, third one skipped

    # Should include 100 from classified data
    first_record = normalizer.get_record(0)
    field_100 = first_record.get_fields("100")[0]
    assert field_100.get_subfields("a")[0] == expected_author
    assert field_100.get_subfields("d")[0] == expected_date
    assert field_100.get_subfields("t")[0] == expected_title
    assert field_100.get_subfields("e")[0] == expected_role

    # Case 2 - existing record updated, original record, should not be changed
    second_record = normalizer.get_record(1)
    field_100 = second_record.get_fields("100")[0]
    # =100  1\$aVilde, Eduard,$d1865-1933.
    assert field_100.get_subfields("a")[0] == "Vilde, Eduard,"
    assert field_100.get_subfields("d")[0] == "1865-1933."


def test_classified_data_with_multiple_records():
    """Test classified data with multiple records - should match by sierraID"""

    classified_fields = [
        [
            {
                "670": {
                    "ind1": " ",
                    "ind2": "0",
                    "subfields": [{"a": "Päikesekiri, 2021"}],
                },
                "111": {
                    "ind1": "2",
                    "ind2": " ",
                    "subfields": [{"a": "Eesti Kirjandusmuuseum"}],
                },
            }
        ],
        [],
        [
            {
                "670": {
                    "ind1": " ",
                    "ind2": "0",
                    "subfields": [{"a": "Teine kirjeldus, 2022"}],
                },
                "111": {
                    "ind1": "2",
                    "ind2": " ",
                    "subfields": [{"a": "ERM"}],
                },
            }
        ],
    ]

    # Case 1 - no 670 exists, should be added to linked record
    for normalizer in (AuthoritiesRecordNormalizer, BibRecordNormalizer):
        linking_results = [
            MOCK_LINKER_NOT_FOUND,
            MOCK_LINKER_ONE_FOUND,
            MOCK_LINKER_NOT_FOUND,
        ]
        normalizer = normalizer(
            linking_results=linking_results, classified_fields=classified_fields
        )

        # Check first record - should have 670 & 111 from classified data
        record = normalizer.first
        assert len(record.get_fields("670")) == 1
        fields_670 = record.get_fields("670")[0]
        fields_111 = record.get_fields("111")[0]
        assert fields_670.get_subfields("a")[0] == "Päikesekiri, 2021"
        assert fields_111.get_subfields("a")[0] == "Eesti Kirjandusmuuseum"

        # Check second record - should not have 670 from classified data
        record = normalizer.get_record(1)
        assert len(record.get_fields("670")) == 1
        fields_670 = record.get_fields("670")[0]
        assert fields_670.get_subfields("a")[0] == "Eesti kirjarahva leksikon, 1995."

        # Check third record - should have 670 from classified data
        record = normalizer.get_record(2)
        assert len(record.get_fields("670")) == 1
        fields_670 = record.get_fields("670")[0]
        assert fields_670.get_subfields("a")[0] == "Teine kirjeldus, 2022"


def test_viaf_name_variations():
    """Test adding alternative name forms from VIAF to 4XX fields. Should skip some variants"""

    normalizer = AuthoritiesRecordNormalizer()
    record = Record()

    viaf_record: VIAFRecord = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=73918842,
        # entity="Jaan Kaplinski"
    )

    assert viaf_record is not None
    assert len(viaf_record.name_variations) > 0

    normalizer._add_author(record, viaf_record)

    fields_4xx = record.get_fields("400", "410", "411")

    unfiltered_name_variations = viaf_record.name_variations

    assert len(fields_4xx) > 0
    assert len(fields_4xx) < len(unfiltered_name_variations)


def test_name_variations_variate_with_author_type():
    """Test that name variations added depend on author type (PER, ORG)"""

    normalizer = AuthoritiesRecordNormalizer()

    # Case 1 - Person
    record = Record()
    viaf_record: VIAFRecord = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=73918842,
        # entity="Jaan Kaplinski"
    )
    normalizer._add_author(record, viaf_record)

    def check_fields_exist(record: Record, tags: list):
        for tag in tags:
            fields = record.get_fields(tag)
            if len(fields) > 0:
                return True
        return False

    # record should only have 400 from 4xx fields
    assert check_fields_exist(record, ["400"])
    assert not check_fields_exist(record, ["410", "411"])

    # Case 2 - Organization
    record = Record()
    viaf_record: VIAFRecord = normalizer._get_viaf_record(
        record,
        False,
        viaf_id=154334700,
        # entity="Eesti Kirjandusmuuseum"
    )
    normalizer._add_author(record, viaf_record)
    # Record should only have 410 from 4xx fields
    assert check_fields_exist(record, ["410"])
    assert not check_fields_exist(record, ["400", "411"])


def test_existing_record_linked_to_viaf_record():
    """Test existing record linked to VIAF record - should enrich with VIAF data"""

    base_path = "tests/test_data/marc_records/json/"
    with (
        open(os.path.join(base_path, "imbi.json"), "r", encoding="utf-8") as f,
        open(os.path.join(base_path, "ernits.json"), "r", encoding="utf-8") as f2,
        open(os.path.join(base_path, "rowling.json"), "r", encoding="utf-8") as f3,
    ):
        imbi = json.load(f)
        ernits = json.load(f2)
        rowling = json.load(f3)

    linking_results = [imbi, ernits, rowling]

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=linking_results,
    )

    def get_viaf_url(record: Record):
        field_024 = record.get_fields("024")
        if len(field_024) == 0:
            return None
        return field_024[0].get_subfields("0")[0]

    viaf_base_url = "http://viaf.org/viaf"
    assert get_viaf_url(normalizer.get_record(0)) == f"{viaf_base_url}/167120147/"
    assert get_viaf_url(normalizer.get_record(1)) == f"{viaf_base_url}/22458146/"
    assert get_viaf_url(normalizer.get_record(1)) == f"{viaf_base_url}/22458146/"
    assert get_viaf_url(normalizer.get_record(2)) == f"{viaf_base_url}/116796842/"


def test_existing_record_enriched_by_elastic_identifier():
    """Test that elastic identifier is correctly parsed from linker response and used to enrich existing record"""

    # Case 1: one linker response with identifier, should find match and enrich
    original_record_json = MOCK_LINKER_ONE_FOUND["linked_info"][0]["json"]
    original_record = next(
        iter(JSONReader(json.dumps([original_record_json], ensure_ascii=False)))
    )
    # check no 024 field with viaf url exists
    assert len(original_record.get_fields("024")) == 0

    linker_res = MOCK_LINKER_ONE_FOUND.copy()
    # add an identifier field to existing record

    normalizer = AuthoritiesRecordNormalizer(
        linking_results=[linker_res],
    )

    record: Record = normalizer.first
    assert len(record.get_fields("024")) == 1
    viaf_url = record.get_fields("024")[0].get_subfields("0")[0]
    expected_viaf_url = "http://viaf.org/viaf/105016/"
    assert viaf_url == expected_viaf_url



from rara_tools.constants.linker  import (
    LOGGER, EntityType
)
from rara_tools.core_formatters.formatted_object import FormattedObject
from typing import List, Dict, NoReturn, Tuple, Any


class FormattedTitle(FormattedObject):
    # TODO: Kas seda on üldse vaja?
    def __init__(self, object_dict: dict, linked_doc: Any):
        super().__init__(
            object_dict=object_dict,
            linked_doc=linked_doc,
            original_entity_key="title"
        )


class FormattedAuthor(FormattedObject):
    def __init__(self, object_dict: dict, linked_doc: Any, entity_type: str):
        super().__init__(
            object_dict=object_dict,
            linked_doc=linked_doc,
            original_entity_key="name"
        )
        self.entity_type: str = entity_type

        self.is_linked: bool = True if self.linked_doc else False # NB! Lisada andmebaasi uus veerg!
        self.original_name: str = self.original_entity            # NB! Lisada andmebaasi uus veerg
        self.author_role: str = self.object_dict.get("role")
        self.is_primary: bool = self.object_dict.get("is_primary")

        self.__primary_author_type: str = None

        self.__name_order_type: str = ""
        self.__event_sub_unit: str = ""
        self.__order_number: str = ""
        self.__sub_title: str = ""
        self.__additional_info: str = ""
        self.__publication_type: str = ""
        self.__publication_language: str = ""
        #self.__standardized_uri: str = ""
        self.__viaf_id: str = ""

        self._default_author_type: str = EntityType.PER


    @property
    def primary_author_type(self) -> str:
        if self.__primary_author_type == None:
            self.__primary_author_type = self._default_author_type
            if self.entity_type != EntityType.UNK:
                if self.entity_type in [EntityType.ORG, EntityType.PER]:
                    self.__primary_author_type = self.entity_type
        return self.__primary_author_type


    @property
    def name(self) -> str:
        """ Force all names into format <last_name>, <first_name>.
        """
        if "," in self.entity:
            name = self.entity 
        else:
            name_tokens = self.entity.rsplit(" ", 1)
            if len(name_tokens) == 2:
                name = f"{name_tokens[1]}, {name_tokens[0]}"
            else:
                name = self.entity 
        return name

    @property
    def name_order(self) -> str:
        if not self.__name_order_type:
            if self.entity_type == EntityType.PER or self.entity_type == EntityType.UNK:
                if "," in self.name:
                    ind1 = "1"
                else:
                    ind1 = "0"
            elif self.entity_type == EntityType.ORG:
                #LOGGER.debug(f"Entity type {self.entity_type} is not {EntityType.PER}.")
                # 1 märksõna esimeseks elemendiks võimupiirkonna nimi, nt:
                #    (a) Eesti (b) Riigikogu - raske automaatselt määrata
                # 2 märksõna esimeseks elemendiks nimi pärijärjestuses
                ind1 = "2" #????????
            else:
                ind1 = "0"
            self.__name_order_type = ind1
        return self.__name_order_type

    @property
    def event_sub_unit(self) -> str:
        if not self.__event_sub_unit:
            self.__event_sub_unit = ""
        return self.__event_sub_unit


    @property
    def order_number(self) -> str:
        if not self.__order_number:
            self.__order_number = ""
        return self.__order_number

    @property
    def sub_title(self) -> str:
        if not self.__sub_title:
            self.__sub_title = ""
        return self.__sub_title

    @property
    def additional_info(self) -> str:
        if not self.__additional_info:
            self.__additional_info = ""
        return self.__additional_info

    @property
    def publication_type(self) -> str:
        if not self.__publication_type:
            self.__publication_type = ""
        return self.__publication_type

    @property
    def publication_language(self) -> str:
        if not self.__publication_language:
            self.__publication_language = ""
        return self.__publication_language

    @property
    def standardized_uri(self) -> str:
        return self.identifier

    @property
    def viaf_id(self):
        if not self.__viaf_id:
            if self.viaf_info:
                self.__viaf_id = self.viaf_info.get("viaf_url", "")
            else:
                self.__viaf_id = ""
        return self.__viaf_id

    def to_dict(self):
        author_dict = {
            "is_linked": self.is_linked,
            "original_name": self.original_name,
            "author_role": self.author_role,
            "is_primary": self.is_primary,
            "primary_author_type": self.primary_author_type,
            "name": self.name,
            "numeration": self.numeration,
            "organisation_sub_unit": self.organisation_sub_unit,
            "titles": self.titles,
            "location": self.location,
            "dates": self.dates,
            "name_order_type": self.name_order,
            "event_sub_unit": self.event_sub_unit,
            "order_number": self.order_number,
            "sub_title": self.sub_title,
            "additional_info": self.additional_info,
            "publication_type": self.publication_type,
            "publication_language": self.publication_language,
            "standardized_uri": self.standardized_uri,
            "viaf_id": self.viaf_id
        }
        return author_dict



import json
import os
import pytest

from time import sleep
from rara_tools.elastic import KataElastic


def load_json(file_path: str):
    with open(file_path, "r") as fh:
        data = json.load(fh)
    return data

TEST_DOCUMENTS = load_json("./tests/test_data/elastic_vectorized_docs.json")
TEST_VECTOR_DATA = load_json("./tests/test_data/test_vector_data.json")
TEST_VECTOR = TEST_VECTOR_DATA.get("vector")
    
es_url = os.getenv("ELASTIC_TEST_URL", "http://localhost:9200")
ELASTIC = KataElastic(es_url)

TEST_KNN_INDEX_NAME = "tools_knn_testing_index"
TEST_ANN_INDEX_NAME = "tools_ann_testing_index"

TEST_VECTOR_FIELD = "vector"



@pytest.mark.order(1)
def test_index_creation_with_knn_vector_mapping():
    """ Tests if index created and documents indexed.
    """
    # Create test index
    created = ELASTIC.create_index(TEST_KNN_INDEX_NAME)
    assert created["acknowledged"] is True
    result = ELASTIC.add_vector_mapping(
        index_name=TEST_KNN_INDEX_NAME, 
        field=TEST_VECTOR_FIELD
    )
    assert result["acknowledged"] is True
    
    
@pytest.mark.order(2)
def test_index_creation_with_ann_vector_mapping():
    """ Tests if index created and documents indexed.
    """
    # Create test index
    created = ELASTIC.create_index(TEST_ANN_INDEX_NAME)
    assert created["acknowledged"] is True
    result = ELASTIC.add_ann_vector_mapping(
        index_name=TEST_ANN_INDEX_NAME, 
        field=TEST_VECTOR_FIELD
    )
    assert result["acknowledged"] is True
    

@pytest.mark.order(3)
def test_vectorized_document_addition_knn_index():
    """ Tests indexing vectorized documents.
    """
    # Add test documents
    for document in TEST_DOCUMENTS:
        indexed = ELASTIC.index_document(TEST_KNN_INDEX_NAME, document)
        assert indexed["result"] == "created"
    # let it index
    sleep(1)
    
@pytest.mark.order(4)
def test_vectorized_document_addition_ann_index():
    """ Tests indexing vectorized documents.
    """
    # Add test documents
    for document in TEST_DOCUMENTS:
        indexed = ELASTIC.index_document(TEST_ANN_INDEX_NAME, document)
        assert indexed["result"] == "created"
    # let it index
    sleep(1)
    
@pytest.mark.order(5)
def test_fuzzy_search():
    """ Tests fuzzy search.
    """
    response = ELASTIC.execute_fuzzy_search(
        index=TEST_ANN_INDEX_NAME,
        field="variations",
        entity="Paul Keres",
        fuzziness=0
    )
    total_hits = response.hits.total.value
    assert total_hits == 2
    
    response = ELASTIC.execute_fuzzy_search(
        index=TEST_ANN_INDEX_NAME,
        field="variations",
        entity="Paul Keres",
        fuzziness=2
    )
    total_hits = response.hits.total.value
    assert total_hits == 3
    
    
@pytest.mark.order(6)
def test_ann_vector_search():
    """ Tests ANN vector search.
    """
    # Execut fuzzy search to get ID restrictions
    response = ELASTIC.execute_fuzzy_search(
        index=TEST_ANN_INDEX_NAME,
        field="variations",
        entity="Paul Keres",
        fuzziness=2
    )
    total_hits = response.hits.total.value
    assert total_hits == 3    
    elastic_ids = [hit.meta.id for hit in response]
    
    response = ELASTIC.execute_ann_vector_search(
        index=TEST_ANN_INDEX_NAME,
        field="vector",
        query_vector=TEST_VECTOR,
        k=1,
        n_docs=1,
        num_candidates=10,
        elastic_ids=elastic_ids
    )
    descriptions = [hit.description for hit in response]
    assert len(descriptions) == 1
    assert descriptions[0] == "Eesti maletaja ja maleteoreetik"

    
@pytest.mark.order(7)
def test_script_score_vector_search():
    """ Tests ANN vector search.
    """
    # Execut fuzzy search to get ID restrictions
    response = ELASTIC.execute_fuzzy_search(
        index=TEST_KNN_INDEX_NAME,
        field="variations",
        entity="Paul Keres",
        fuzziness=2
    )
    total_hits = response.hits.total.value
    assert total_hits == 3    
    elastic_ids = [hit.meta.id for hit in response]
    
    response = ELASTIC.execute_script_score_vector_search(
        index=TEST_KNN_INDEX_NAME,
        field="vector",
        query_vector=TEST_VECTOR,
        n_docs=1,
        elastic_ids=elastic_ids
    )
    descriptions = [hit.description for hit in response]
    assert len(descriptions) == 1
    assert descriptions[0] == "Eesti maletaja ja maleteoreetik"

    
@pytest.mark.order(8)
def test_index_deleting():
    """
    Tests deleting index. We delete the test index now.
    """
    indices = [TEST_KNN_INDEX_NAME, TEST_ANN_INDEX_NAME]
    for index in indices:
        deleted = ELASTIC.delete_index(index)
        sleep(1)
        assert deleted["acknowledged"] is True
        



COMPONENT_KEY = "digitizer"
ES_INDEX_DEFAULT = "digitizer_output"


class ModelTypes:
    IMAGE_PROCESSOR = "image_processor"


class StatusKeys:
    DOWNLOAD_MODELS = "digitizer_download_models"
    CLEAN_UP = "digitizer_clean_up"
    ELASTICSEARCH_UPLOAD = "digitizer_elasticsearch_upload"
    UPLOAD = "s3_upload"
    DOWNLOAD = "digitizer_s3_download"
    OCR = "digitizer_ocr"


class Queue:
    IO = "io"
    DOWNLOAD = "download"
    FINISH = "finish"
    OCR = "ocr"
    UTILITY = "digitizer-utility"


class Tasks:
    START_DIGITIZER_PIPELINE = "start_digitizer_pipeline"
    PURGE_MODELS = "purge_unused_digitizer_models"


class Error:
    NO_SPACE = "Disk out of space!"
    COULDNT_DOWNLOAD = "Unknown error when downloading model!"
    UNKNOWN = "Unknown system error!"
    S3_CONNECTION = "Failed to connect to S3!"
    UNSUPPORTED_FILETYPE = "Unsupported file type!"
    COULDNT_UPLOAD = "Could not upload documents to Elasticsearch!"
    FILE_IS_PROTECTED = "File is password protected or encrypted!"
    UNKNOWN_OCR = "Unknown error when applying ocr!"
    CUSTOM_MODEL_ERROR = "Couldn't download custom image classification model!"
